/*
 * arith.c -- arithmetic routines
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "common.h"
#include "arith.h"


#define ROUND_OFF 1
#define ROUND_UP 2
#define ROUND_DOWN 3
#define ROUND_NEAR 4

#define X_BIT (1<<8) 
#define Z_BIT (1<<9) 
#define U_BIT (1<<10) 
#define O_BIT (1<<11) 
#define I_BIT (1<<12) 
#define W_BIT (1<<13) 
#define V_BIT (1<<14) 
#define D_BIT (1<<15) 
#define E_BIT (1<<18)

#define zero_exponent (-1000)

#define bignum_prec 157

#define magic_offset 2112
#define origin 37

#define buf0 (buf+8) 
#define buf_max (buf+777)


typedef enum {
  zro, num, inf, nan
} ftype;

typedef struct {
  int a;
  int b;
  Tetra dat[bignum_prec];
} bignum;


static void bignum_times_ten(bignum *f);
static int bignum_compare(bignum *f, bignum *g);
static void bignum_dec(bignum *f, bignum *g, Tetra r);
static void bignum_double(bignum *f);


Octa zero = { 0, 0 };
Octa negOne = { -1, -1 };
Octa infOcta = { 0x7ff00000, 0 };
Octa standardNaN = { 0x7ff80000, 0 };
Octa aux;
Bool overflow;
int curRound;
int exceptions;

Octa val;
char *next_char;

static char buf[785] = "00000000";


Octa oplus(Octa y, Octa z) {
  Octa x;

  x.h = y.h + z.h;
  x.l = y.l + z.l;
  if (x.l < y.l) x.h++;
  return x;
}


Octa ominus(Octa y, Octa z) {
  Octa x;

  x.h = y.h - z.h;
  x.l = y.l - z.l;
  if (x.l > y.l) x.h--;
  return x;
}


Octa incr(Octa y, int delta) {
  Octa x;

  x.h = y.h;
  x.l = y.l + delta;
  if (delta >= 0) {
    if (x.l < y.l) x.h++;
  } else
    if (x.l > y.l) x.h--;
  return x;
}


Octa shiftLeft(y,s)
Octa y;
int s;
{
while(s>=32)y.h= y.l,y.l= 0,s-= 32;
if(s){register Tetra yhl= y.h<<s,ylh= y.l>>(32-s);
y.h= yhl+ylh;y.l<<= s;
}
return y;
}


Octa shiftRight(Octa y, int s, Bool u) {
  while (s >= 32) y.l = y.h, y.h = (u ? 0 : -(y.h>>31)), s-= 32;
  if (s) {
    Tetra yhl = y.h << (32 - s), ylh = y.l >> s;
    y.h = (u ? 0 : (-(y.h >> 31)) << (32-s)) + (y.h >> s);
    y.l = yhl + ylh;
  }
  return y;
}


Octa omult(y,z)
Octa y,z;
{
register int i,j,k;
Tetra u[4],v[4],w[8];
register Tetra t;
Octa acc;

u[3]= y.h>>16,u[2]= y.h&0xffff,u[1]= y.l>>16,u[0]= y.l&0xffff;
v[3]= z.h>>16,v[2]= z.h&0xffff,v[1]= z.l>>16,v[0]= z.l&0xffff;

for(j= 0;j<4;j++)w[j]= 0;
for(j= 0;j<4;j++)
if(!v[j])w[j+4]= 0;
else{
for(i= k= 0;i<4;i++){
t= u[i]*v[j]+w[i+j]+k;
w[i+j]= t&0xffff,k= t>>16;
}
w[j+4]= k;
}

aux.h= (w[7]<<16)+w[6],aux.l= (w[5]<<16)+w[4];
acc.h= (w[3]<<16)+w[2],acc.l= (w[1]<<16)+w[0];

return acc;
}


Octa somult(y,z)
Octa y,z;
{
Octa acc;
acc= omult(y,z);
if(y.h&SIGN_BIT)aux= ominus(aux,z);
if(z.h&SIGN_BIT)aux= ominus(aux,y);
overflow= (aux.h!=aux.l||(aux.h^(aux.h>>1)^(acc.h&SIGN_BIT)));
return acc;
}


Octa odiv(x,y,z)
Octa x,y,z;
{
register int i,j,k,n,d;
Tetra u[8],v[4],q[4],mask,qhat,rhat,vh,vmh;
register Tetra t;
Octa acc;
/*14:*/

if(x.h> z.h||(x.h==z.h&&x.l>=z.l)){
aux= y;return x;
}

/*:14*/
;
/*15:*/

u[7]= x.h>>16,u[6]= x.h&0xffff,u[5]= x.l>>16,u[4]= x.l&0xffff;
u[3]= y.h>>16,u[2]= y.h&0xffff,u[1]= y.l>>16,u[0]= y.l&0xffff;
v[3]= z.h>>16,v[2]= z.h&0xffff,v[1]= z.l>>16,v[0]= z.l&0xffff;

/*:15*/
;
/*16:*/

for(n= 4;v[n-1]==0;n--);

/*:16*/
;
/*17:*/

vh= v[n-1];
for(d= 0;vh<0x8000;d++,vh<<= 1);
for(j= k= 0;j<n+4;j++){
t= (u[j]<<d)+k;
u[j]= t&0xffff,k= t>>16;
}
for(j= k= 0;j<n;j++){
t= (v[j]<<d)+k;
v[j]= t&0xffff,k= t>>16;
}
vh= v[n-1];
vmh= (n> 1?v[n-2]:0);

/*:17*/
;
for(j= 3;j>=0;j--)/*20:*/

{
/*21:*/

t= (u[j+n]<<16)+u[j+n-1];
qhat= t/vh,rhat= t-vh*qhat;
if(n> 1)while(qhat==0x10000||qhat*vmh> (rhat<<16)+u[j+n-2]){
qhat--,rhat+= vh;
if(rhat>=0x10000)break;
}

/*:21*/
;
/*22:*/

for(i= k= 0;i<n;i++){
t= u[i+j]+0xffff0000-k-qhat*v[i];
u[i+j]= t&0xffff,k= 0xffff-(t>>16);
}

/*:22*/
;
/*23:*/

if(u[j+n]!=k){
qhat--;
for(i= k= 0;i<n;i++){
t= u[i+j]+v[i]+k;
u[i+j]= t&0xffff,k= t>>16;
}
}

/*:23*/
;
q[j]= qhat;
}

/*:20*/
;
/*18:*/

mask= (1<<d)-1;
for(j= 3;j>=n;j--)u[j]= 0;
for(k= 0;j>=0;j--){
t= (k<<16)+u[j];
u[j]= t>>d,k= t&mask;
}

/*:18*/
;
/*19:*/

acc.h= (q[3]<<16)+q[2],acc.l= (q[1]<<16)+q[0];
aux.h= (u[3]<<16)+u[2],aux.l= (u[1]<<16)+u[0];

/*:19*/
;
return acc;
}

/*:13*//*24:*/

Octa sodiv(y,z)
Octa y,z;
{
Octa yy,zz,q;
register int sy,sz;
if(y.h&SIGN_BIT)sy= 2,yy= ominus(zero,y);
else sy= 0,yy= y;
if(z.h&SIGN_BIT)sz= 1,zz= ominus(zero,z);
else sz= 0,zz= z;
q= odiv(zero,yy,zz);
overflow= false;
  switch (sy+sz) {
    case 2+1:
      aux= ominus(zero,aux);
      if (q.h==SIGN_BIT) overflow= true;
    case 0+0:
      return q;
    case 2+0:
      if(aux.h||aux.l)aux= ominus(zz,aux);
      goto negate_q;
    case 0+1:
      if (aux.h||aux.l) aux= ominus(aux,zz);
negate_q:
      if (aux.h||aux.l) return ominus(negOne,q);
      else return ominus(zero,q);
  }
  fprintf(stderr,
          "Error: this should not happen (signed_odiv)\n");
  exit(1);
  return zero;
}

/*:24*//*25:*/

Octa oand(y,z)
Octa y,z;
{Octa x;
x.h= y.h&z.h;x.l= y.l&z.l;
return x;
}

Octa oandn(y,z)
Octa y,z;
{Octa x;
x.h= y.h&~z.h;x.l= y.l&~z.l;
return x;
}

Octa oxor(y,z)
Octa y,z;
{Octa x;
x.h= y.h^z.h;x.l= y.l^z.l;
return x;
}

/*:25*//*26:*/

int countBits(Tetra x) {
register int xx= x;
xx= xx-((xx>>1)&0x55555555);
xx= (xx&0x33333333)+((xx>>2)&0x33333333);
xx= (xx+(xx>>4))&0x0f0f0f0f;
xx= xx+(xx>>8);
return(xx+(xx>>16))&0xff;
}

/*:26*//*27:*/

Tetra byteDiff(y,z)
Tetra y,z;
{
register Tetra d= (y&0x00ff00ff)+0x01000100-(z&0x00ff00ff);
register Tetra m= d&0x01000100;
register Tetra x= d&(m-(m>>8));
d= ((y>>8)&0x00ff00ff)+0x01000100-((z>>8)&0x00ff00ff);
m= d&0x01000100;
return x+((d&(m-(m>>8)))<<8);
}

/*:27*//*28:*/

Tetra wydeDiff(y,z)
Tetra y,z;
{
register Tetra a= ((y>>16)-(z>>16))&0x10000;
register Tetra b= ((y&0xffff)-(z&0xffff))&0x10000;
return y-(z^((y^z)&(b-a-(b>>16))));
}

/*:28*//*29:*/

Octa boolMult(y,z,xor)
Octa y,z;
Bool xor;
{
Octa o,x;
register Tetra a,b,c;
register int k;
for(k= 0,o= y,x= zero;o.h||o.l;k++,o= shiftRight(o,8,1))
if(o.l&0xff){
a= ((z.h>>k)&0x01010101)*0xff;
b= ((z.l>>k)&0x01010101)*0xff;
c= (o.l&0xff)*0x01010101;
if(xor)x.h^= a&c,x.l^= b&c;
else x.h|= a&c,x.l|= b&c;
}
return x;
}

/*:29*//*31:*/

Octa fpack(f,e,s,r)
Octa f;
int e;
char s;
int r;
{
Octa o;
if(e> 0x7fd)e= 0x7ff,o= zero;
else{
if(e<0){
if(e<-54)o.h= 0,o.l= 1;
else{Octa oo;
o= shiftRight(f,-e,1);
oo= shiftLeft(o,-e);
if(oo.l!=f.l||oo.h!=f.h)o.l|= 1;

}
e= 0;
}else o= f;
}
/*33:*/

if(o.l&3)exceptions|= X_BIT;
switch(r){
case ROUND_DOWN:if(s=='-')o= incr(o,3);break;
case ROUND_UP:if(s!='-')o= incr(o,3);
case ROUND_OFF:break;
case ROUND_NEAR:o= incr(o,o.l&4?2:1);break;
}
o= shiftRight(o,2,1);
o.h+= e<<20;
if(o.h>=0x7ff00000)exceptions|= O_BIT+X_BIT;
else if(o.h<0x100000)exceptions|= U_BIT;
if(s=='-')o.h|= SIGN_BIT;
return o;

/*:33*/
;
}

/*:31*//*34:*/

Tetra sfpack(f,e,s,r)
Octa f;
int e;
char s;
int r;
{
register Tetra o;
if(e> 0x47d)e= 0x47f,o= 0;
else{
o= shiftLeft(f,3).h;
if(f.l&0x1fffffff)o|= 1;
if(e<0x380){
if(e<0x380-25)o= 1;
else{register Tetra o0,oo;
o0= o;
o= o>>(0x380-e);
oo= o<<(0x380-e);
if(oo!=o0)o|= 1;

}
e= 0x380;
}
}
/*35:*/

if(o&3)exceptions|= X_BIT;
switch(r){
case ROUND_DOWN:if(s=='-')o+= 3;break;
case ROUND_UP:if(s!='-')o+= 3;
case ROUND_OFF:break;
case ROUND_NEAR:o+= (o&4?2:1);break;
}
o= o>>2;
o+= (e-0x380)<<23;
if(o>=0x7f800000)exceptions|= O_BIT+X_BIT;
else if(o<0x100000)exceptions|= U_BIT;
if(s=='-')o|= SIGN_BIT;
return o;

/*:35*/
;
}

/*:34*//*37:*/

ftype funpack(x,f,e,s)
Octa x;
Octa*f;
int*e;
char*s;
{
register int ee;
exceptions= 0;
*s= (x.h&SIGN_BIT?'-':'+');
*f= shiftLeft(x,2);
f->h&= 0x3fffff;
ee= (x.h>>20)&0x7ff;
if(ee){
*e= ee-1;
f->h|= 0x400000;
return(ee<0x7ff?num:f->h==0x400000&&!f->l?inf:nan);
}
if(!x.l&&!f->h){
*e= zero_exponent;return zro;
}
do{ee--;*f= shiftLeft(*f,1);}while(!(f->h&0x400000));
*e= ee;return num;
}

/*:37*//*38:*/

ftype sfunpack(x,f,e,s)
Tetra x;
Octa*f;
int*e;
char*s;
{
register int ee;
exceptions= 0;
*s= (x&SIGN_BIT?'-':'+');
f->h= (x>>1)&0x3fffff,f->l= x<<31;
ee= (x>>23)&0xff;
if(ee){
*e= ee+0x380-1;
f->h|= 0x400000;
return(ee<0xff?num:(x&0x7fffffff)==0x7f800000?inf:nan);
}
if(!(x&0x7fffffff)){
*e= zero_exponent;return zro;
}
do{ee--;*f= shiftLeft(*f,1);}while(!(f->h&0x400000));
*e= ee+0x380;return num;
}

/*:38*//*39:*/

Octa sfLoad(z)
Tetra z;
{
Octa f,x;int e;char s;ftype t;
t= sfunpack(z,&f,&e,&s);
switch(t){
case zro:x= zero;break;
case num:return fpack(f,e,s,ROUND_OFF);
case inf:x= infOcta;break;
case nan:x= shiftRight(f,2,1);x.h|= 0x7ff00000;break;
}
if(s=='-')x.h|= SIGN_BIT;
return x;
}

/*:39*//*40:*/

Tetra sfStore(x)
Octa x;
{
Octa f;Tetra z;int e;char s;ftype t;
t= funpack(x,&f,&e,&s);
switch(t){
case zro:z= 0;break;
case num:return sfpack(f,e,s,curRound);
case inf:z= 0x7f800000;break;
case nan:if(!(f.h&0x200000)){
f.h|= 0x200000;exceptions|= I_BIT;
}
z= 0x7f800000|(f.h<<1)|(f.l>>31);break;
}
if(s=='-')z|= SIGN_BIT;
return z;
}

/*:40*//*41:*/

Octa fmult(y,z)
Octa y,z;
{
ftype yt,zt;
int ye,ze;
char ys,zs;
Octa x,xf,yf,zf;
register int xe;
register char xs;
yt= funpack(y,&yf,&ye,&ys);
zt= funpack(z,&zf,&ze,&zs);
xs= ys+zs-'+';
switch(4*yt+zt){
/*42:*/

case 4*nan+nan:if(!(y.h&0x80000))exceptions|= I_BIT;
case 4*zro+nan:case 4*num+nan:case 4*inf+nan:
if(!(z.h&0x80000))exceptions|= I_BIT,z.h|= 0x80000;
return z;
case 4*nan+zro:case 4*nan+num:case 4*nan+inf:
if(!(y.h&0x80000))exceptions|= I_BIT,y.h|= 0x80000;
return y;

/*:42*/
;
case 4*zro+zro:case 4*zro+num:case 4*num+zro:x= zero;break;
case 4*num+inf:case 4*inf+num:case 4*inf+inf:x= infOcta;break;
case 4*zro+inf:case 4*inf+zro:x= standardNaN;
exceptions|= I_BIT;break;
case 4*num+num:/*43:*/

xe= ye+ze-0x3fd;
x= omult(yf,shiftLeft(zf,9));
if(aux.h>=0x400000)xf= aux;
else xf= shiftLeft(aux,1),xe--;
if(x.h||x.l)xf.l|= 1;
return fpack(xf,xe,xs,curRound);

/*:43*/
;
}
if(xs=='-')x.h|= SIGN_BIT;
return x;
}

/*:41*//*44:*/

Octa fdivide(y,z)
Octa y,z;
{
ftype yt,zt;
int ye,ze;
char ys,zs;
Octa x,xf,yf,zf;
register int xe;
register char xs;
yt= funpack(y,&yf,&ye,&ys);
zt= funpack(z,&zf,&ze,&zs);
xs= ys+zs-'+';
switch(4*yt+zt){
/*42:*/

case 4*nan+nan:if(!(y.h&0x80000))exceptions|= I_BIT;
case 4*zro+nan:case 4*num+nan:case 4*inf+nan:
if(!(z.h&0x80000))exceptions|= I_BIT,z.h|= 0x80000;
return z;
case 4*nan+zro:case 4*nan+num:case 4*nan+inf:
if(!(y.h&0x80000))exceptions|= I_BIT,y.h|= 0x80000;
return y;

/*:42*/
;
case 4*zro+inf:case 4*zro+num:case 4*num+inf:x= zero;break;
case 4*num+zro:exceptions|= Z_BIT;
case 4*inf+num:case 4*inf+zro:x= infOcta;break;
case 4*zro+zro:case 4*inf+inf:x= standardNaN;
exceptions|= I_BIT;break;
case 4*num+num:/*45:*/

xe= ye-ze+0x3fd;
xf= odiv(yf,zero,shiftLeft(zf,9));
if(xf.h>=0x800000){
aux.l|= xf.l&1;
xf= shiftRight(xf,1,1);
xe++;
}
if(aux.h||aux.l)xf.l|= 1;
return fpack(xf,xe,xs,curRound);

/*:45*/
;
}
if(xs=='-')x.h|= SIGN_BIT;
return x;
}

/*:44*//*46:*/

Octa fplus(y,z)
Octa y,z;
{
ftype yt,zt;
int ye,ze;
char ys,zs;
Octa x,xf,yf,zf;
register int xe,d;
register char xs = '\0';
yt= funpack(y,&yf,&ye,&ys);
zt= funpack(z,&zf,&ze,&zs);
switch(4*yt+zt){
/*42:*/

case 4*nan+nan:if(!(y.h&0x80000))exceptions|= I_BIT;
case 4*zro+nan:case 4*num+nan:case 4*inf+nan:
if(!(z.h&0x80000))exceptions|= I_BIT,z.h|= 0x80000;
return z;
case 4*nan+zro:case 4*nan+num:case 4*nan+inf:
if(!(y.h&0x80000))exceptions|= I_BIT,y.h|= 0x80000;
return y;

/*:42*/
;
case 4*zro+num:return fpack(zf,ze,zs,ROUND_OFF);break;
case 4*num+zro:return fpack(yf,ye,ys,ROUND_OFF);break;
case 4*inf+inf:if(ys!=zs){
exceptions|= I_BIT;x= standardNaN;xs= zs;break;
}
case 4*num+inf:case 4*zro+inf:x= infOcta;xs= zs;break;
case 4*inf+num:case 4*inf+zro:x= infOcta;xs= ys;break;
case 4*num+num:if(y.h!=(z.h^0x80000000)||y.l!=z.l)
/*47:*/

{Octa o,oo;
if(ye<ze||(ye==ze&&(yf.h<zf.h||(yf.h==zf.h&&yf.l<zf.l))))
/*48:*/

{
o= yf,yf= zf,zf= o;
d= ye,ye= ze,ze= d;
d= ys,ys= zs,zs= d;
}

/*:48*/
;
d= ye-ze;
xs= ys,xe= ye;
if(d)/*49:*/

{
if(d<=2)zf= shiftRight(zf,d,1);
else if(d> 53)zf.h= 0,zf.l= 1;
else{
if(ys!=zs)d--,xe--,yf= shiftLeft(yf,1);
o= zf;
zf= shiftRight(o,d,1);
oo= shiftLeft(zf,d);
if(oo.l!=o.l||oo.h!=o.h)zf.l|= 1;
}
}

/*:49*/
;
if(ys==zs){
xf= oplus(yf,zf);
if(xf.h>=0x800000)xe++,d= xf.l&1,xf= shiftRight(xf,1,1),xf.l|= d;
}else{
xf= ominus(yf,zf);
if(xf.h>=0x800000)xe++,d= xf.l&1,xf= shiftRight(xf,1,1),xf.l|= d;
else while(xf.h<0x400000)xe--,xf= shiftLeft(xf,1);
}
return fpack(xf,xe,xs,curRound);
}

/*:47*/
;
case 4*zro+zro:x= zero;
xs= (ys==zs?ys:curRound==ROUND_DOWN?'-':'+');break;
}
if(xs=='-')x.h|= SIGN_BIT;
return x;
}

/*:46*//*50:*/

int fepscomp(y,z,e,s)
Octa y,z,e;
int s;
{
Octa yf,zf,ef,o,oo;
int ye,ze,ee;
char ys,zs,es;
register int yt,zt,et,d;
et= funpack(e,&ef,&ee,&es);
if(es=='-')return 2;
switch(et){
case nan:return 2;
case inf:ee= 10000;
case num:case zro:break;
}
yt= funpack(y,&yf,&ye,&ys);
zt= funpack(z,&zf,&ze,&zs);
switch(4*yt+zt){
case 4*nan+nan:case 4*nan+inf:case 4*nan+num:case 4*nan+zro:
case 4*inf+nan:case 4*num+nan:case 4*zro+nan:return 2;
case 4*inf+inf:return(ys==zs||ee>=1023);
case 4*inf+num:case 4*inf+zro:case 4*num+inf:case 4*zro+inf:
return(s&&ee>=1022);
case 4*zro+zro:return 1;
case 4*zro+num:case 4*num+zro:if(!s)return 0;
case 4*num+num:break;
}
/*51:*/

/*52:*/

if(ye<0)yf= shiftLeft(y,2),ye= 0;
if(ze<0)zf= shiftLeft(z,2),ze= 0;

/*:52*/
;
if(ye<ze||(ye==ze&&(yf.h<zf.h||(yf.h==zf.h&&yf.l<zf.l))))
/*48:*/

{
o= yf,yf= zf,zf= o;
d= ye,ye= ze,ze= d;
d= ys,ys= zs,zs= d;
}

/*:48*/
;
if(ze==zero_exponent)ze= ye;
d= ye-ze;
if(!s)ee-= d;
if(ee>=1023)return 1;
/*53:*/

if(d> 54)o= zero,oo= zf;
else o= shiftRight(zf,d,1),oo= shiftLeft(o,d);
if(oo.h!=zf.h||oo.l!=zf.l){
if(ee<1020)return 0;
o= incr(o,ys==zs?0:1);
}
o= (ys==zs?ominus(yf,o):oplus(yf,o));

/*:53*/
;
if(!o.h&&!o.l)return 1;
if(ee<968)return 0;
if(ee>=1021)ef= shiftLeft(ef,ee-1021);
else ef= shiftRight(ef,1021-ee,1);
return o.h<ef.h||(o.h==ef.h&&o.l<=ef.l);

/*:51*/
;
}

/*:50*//*54:*/

void print_float(x)
Octa x;
{
/*56:*/

Octa f,g;
register int e;
register int j,k;

/*:56*//*66:*/

bignum ff,gg;
bignum tt;
char s[18];
register char*p;

/*:66*/
;
if(x.h&SIGN_BIT)printf("-");
/*55:*/

f= shiftLeft(x,1);
e= f.h>>21;
f.h&= 0x1fffff;
if(!f.h&&!f.l)/*57:*/

{
if(!e){
printf("0.");return;
}
if(e==0x7ff){
printf("Inf");return;
}
e--;
f.h= 0x3fffff,f.l= 0xffffffff;
g.h= 0x400000,g.l= 2;
}

/*:57*/

else{
g= incr(f,1);
f= incr(f,-1);
if(!e)e= 1;
else if(e==0x7ff){
printf("NaN");
if(g.h==0x100000&&g.l==1)return;
e= 0x3ff;
}else f.h|= 0x200000,g.h|= 0x200000;
}

/*:55*/
;
/*63:*/

k= (magic_offset-e)/28;
ff.dat[k-1]= shiftRight(f,magic_offset+28-e-28*k,1).l&0xfffffff;
gg.dat[k-1]= shiftRight(g,magic_offset+28-e-28*k,1).l&0xfffffff;
ff.dat[k]= shiftRight(f,magic_offset-e-28*k,1).l&0xfffffff;
gg.dat[k]= shiftRight(g,magic_offset-e-28*k,1).l&0xfffffff;
ff.dat[k+1]= shiftLeft(f,e+28*k-(magic_offset-28)).l&0xfffffff;
gg.dat[k+1]= shiftLeft(g,e+28*k-(magic_offset-28)).l&0xfffffff;
ff.a= (ff.dat[k-1]?k-1:k);
ff.b= (ff.dat[k+1]?k+1:k);
gg.a= (gg.dat[k-1]?k-1:k);
gg.b= (gg.dat[k+1]?k+1:k);

/*:63*/
;
/*64:*/

if(e> 0x401)/*65:*/

{register int open= x.l&1;
tt.dat[origin]= 10;
tt.a= tt.b= origin;
for(e= 1;bignum_compare(&gg,&tt)>=open;e++)
bignum_times_ten(&tt);
p= s;
while(1){
bignum_times_ten(&ff);
bignum_times_ten(&gg);
for(j= '0';bignum_compare(&ff,&tt)>=0;j++)
bignum_dec(&ff,&tt,0x10000000),bignum_dec(&gg,&tt,0x10000000);
if(bignum_compare(&gg,&tt)>=open)break;
*p++= j;
if(ff.a==bignum_prec-1&&!open)
goto done;
}
for(k= j;bignum_compare(&gg,&tt)>=open;k++)bignum_dec(&gg,&tt,0x10000000);
*p++= (j+1+k)>>1;
done:;
}

/*:65*/

else{
if(ff.a> origin)ff.dat[origin]= 0;
for(e= 1,p= s;gg.a> origin||ff.dat[origin]==gg.dat[origin];){
if(gg.a> origin)e--;
else*p++= ff.dat[origin]+'0',ff.dat[origin]= 0,gg.dat[origin]= 0;
bignum_times_ten(&ff);
bignum_times_ten(&gg);
}
*p++= ((ff.dat[origin]+1+gg.dat[origin])>>1)+'0';
}
*p= '\0';

/*:64*/
;
/*67:*/

if(e> 17||e<(int)strlen(s)-17)
printf("%c%s%se%d",s[0],(s[1]?".":""),s+1,e-1);
else if(e<0)printf(".%0*d%s",-e,0,s);
else if(strlen(s)>=e)printf("%.*s.%s",e,s,s+e);
else printf("%s%0*d.",s,e-(int)strlen(s),0);

/*:67*/
;
}

/*:54*//*60:*/

static void bignum_times_ten(bignum *f) {
register Tetra*p,*q;register Tetra x,carry;
for(p= &f->dat[f->b],q= &f->dat[f->a],carry= 0;p>=q;p--){
x= *p*10+carry;
*p= x&0xfffffff;
carry= x>>28;
}
*p= carry;
if(carry)f->a--;
if(f->dat[f->b]==0&&f->b> f->a)f->b--;
}

/*:60*//*61:*/

static int bignum_compare(bignum *f, bignum *g) {
register Tetra*p,*pp,*q,*qq;
if(f->a!=g->a)return f->a> g->a?-1:1;
pp= &f->dat[f->b],qq= &g->dat[g->b];
for(p= &f->dat[f->a],q= &g->dat[g->a];p<=pp;p++,q++){
if(*p!=*q)return*p<*q?-1:1;
if(q==qq)return p<pp;
}
return-1;
}

/*:61*//*62:*/

static void bignum_dec(bignum *f, bignum *g, Tetra r) {
register Tetra*p,*q,*qq;
register int x,borrow;
while(g->b> f->b)f->dat[++f->b]= 0;
qq= &g->dat[g->a];
for(p= &f->dat[g->b],q= &g->dat[g->b],borrow= 0;q>=qq;p--,q--){
x= *p-*q-borrow;
if(x>=0)borrow= 0,*p= x;
else borrow= 1,*p= x+r;
}
for(;borrow;p--)
if(*p)borrow= 0,*p= *p-1;
else*p= r-1;
while(f->dat[f->a]==0){
if(f->a==f->b){
f->a= f->b= bignum_prec-1,f->dat[bignum_prec-1]= 0;
return;
}
f->a++;
}
while(f->dat[f->b]==0)f->b--;
}

/*:62*//*68:*/

int scan_const(s)
char*s;
{
/*70:*/

register char*p,*q;
register Bool NaN;
int sign;

/*:70*//*76:*/

register char*dec_pt;
register int exp = 0;
register int zeros = 0;

/*:76*//*81:*/

register int k,x;
register char*pp;
bignum ff,tt;

/*:81*/
;
val.h= val.l= 0;
p= s;
if(*p=='+'||*p=='-')sign= *p++;else sign= '+';
if(strncmp(p,"NaN",3)==0)NaN= true,p+= 3;
else NaN= false;
if((isdigit(*p)&&!NaN)||(*p=='.'&&isdigit(*(p+1))))
/*73:*/

{
  for (q= buf0,dec_pt= (char*)0;isdigit(*p);p++) {
    val= oplus(val,shiftLeft(val,2));
    val= incr(shiftLeft(val,1),*p-'0');
    if (q> buf0||*p!='0') {
      if (q<buf_max) {
        *q++= *p;
      } else {
        if (*(q-1)=='0') {
          *(q-1)= *p;
        }
      }
    }
  }
if(NaN)*q++= '1';
if(*p=='.')/*74:*/

{
dec_pt= q;
p++;
for(zeros= 0;isdigit(*p);p++)
if(*p=='0'&&q==buf0)zeros++;
else if(q<buf_max)*q++= *p;
else if(*(q-1)=='0')*(q-1)= *p;
}

/*:74*/
;
next_char= p;
if(*p=='e'&&!NaN)/*77:*/

{register char exp_sign;
p++;
if(*p=='+'||*p=='-')exp_sign= *p++;else exp_sign= '+';
if(isdigit(*p)){
for(exp= *p++-'0';isdigit(*p);p++)
if(exp<1000)exp= 10*exp+*p-'0';
if(!dec_pt)dec_pt= q,zeros= 0;
if(exp_sign=='-')exp= -exp;
next_char= p;
}
}

/*:77*/

else exp= 0;
if(dec_pt)/*78:*/

{
/*79:*/

x= buf+341+zeros-dec_pt-exp;
if(q==buf0||x>=1413){
/* make_it_zero: */ exp= -99999;goto packit;
}
if(x<0){
make_it_infinite:exp= 99999;goto packit;
}
ff.a= x/9;
for(p= q;p<q+8;p++)*p= '0';
q= q-1-(q+341+zeros-dec_pt-exp)%9;
for(p= buf0-x%9,k= ff.a;p<=q&&k<=156;p+= 9,k++)
/*80:*/

{
for(x= *p-'0',pp= p+1;pp<p+9;pp++)x= 10*x+*pp-'0';
ff.dat[k]= x;
}

/*:80*/
;
ff.b= k-1;
for(x= 0;p<=q;p+= 9)if(strncmp(p,"000000000",9)!=0)x= 1;
ff.dat[156]+= x;

while(ff.dat[ff.b]==0)ff.b--;

/*:79*/
;
/*83:*/

val= zero;
if(ff.a> 36){
for(exp= 0x3fe;ff.a> 36;exp--)bignum_double(&ff);
for(k= 54;k;k--){
if(ff.dat[36]){
if(k>=32)val.h|= 1<<(k-32);else val.l|= 1<<k;
ff.dat[36]= 0;
if(ff.b==36)break;
}
bignum_double(&ff);
}
}else{
tt.a= tt.b= 36,tt.dat[36]= 2;
for(exp= 0x3fe;bignum_compare(&ff,&tt)>=0;exp++)bignum_double(&tt);
for(k= 54;k;k--){
bignum_double(&ff);
if(bignum_compare(&ff,&tt)>=0){
if(k>=32)val.h|= 1<<(k-32);else val.l|= 1<<k;
bignum_dec(&ff,&tt,1000000000);
if(ff.a==bignum_prec-1)break;
}
}
}
if(k==0)val.l|= 1;

/*:83*/
;
packit:/*84:*/

val= fpack(val,exp,sign,ROUND_NEAR);
if(NaN){
if((val.h&0x7fffffff)==0x40000000)val.h|= 0x7fffffff,val.l= 0xffffffff;
else if((val.h&0x7fffffff)==0x3ff00000&&!val.l)val.h|= 0x40000000,val.l= 1;
else val.h|= 0x40000000;
}

/*:84*/
;
return 1;
}

/*:78*/
;
if(sign=='-')val= ominus(zero,val);
return 0;
}

/*:73*/
;
if(NaN)/*71:*/

{
next_char= p;
val.h= 0x600000,exp= 0x3fe;
goto packit;
}

/*:71*/
;
if(strncmp(p,"Inf",3)==0)/*72:*/

{
next_char= p+3;
goto make_it_infinite;
}

/*:72*/
;
/* no_const_found: */ next_char= s;return-1;
}

/*:68*//*82:*/

static void bignum_double(bignum *f) {
register Tetra*p,*q;register int x,carry;
for(p= &f->dat[f->b],q= &f->dat[f->a],carry= 0;p>=q;p--){
x= *p+*p+carry;
if(x>=1000000000)carry= 1,*p= x-1000000000;
else carry= 0,*p= x;
}
*p= carry;
if(carry)f->a--;
if(f->dat[f->b]==0&&f->b> f->a)f->b--;
}

/*:82*//*85:*/

int fcomp(y,z)
Octa y,z;
{
ftype yt,zt;
int ye,ze;
char ys,zs;
Octa yf,zf;
register int x = 0;
yt= funpack(y,&yf,&ye,&ys);
zt= funpack(z,&zf,&ze,&zs);
switch(4*yt+zt){
case 4*nan+nan:case 4*zro+nan:case 4*num+nan:case 4*inf+nan:
case 4*nan+zro:case 4*nan+num:case 4*nan+inf:return 2;
case 4*zro+zro:return 0;
case 4*zro+num:case 4*num+zro:case 4*zro+inf:case 4*inf+zro:
case 4*num+num:case 4*num+inf:case 4*inf+num:case 4*inf+inf:
if(ys!=zs)x= 1;
else if(y.h> z.h)x= 1;
else if(y.h<z.h)x= -1;
else if(y.l> z.l)x= 1;
else if(y.l<z.l)x= -1;
else return 0;
break;
}
return(ys=='-'?-x:x);
}

/*:85*//*86:*/

Octa fintegerize(z,r)
Octa z;
int r;
{
ftype zt;
int ze;
char zs;
Octa xf,zf;
zt= funpack(z,&zf,&ze,&zs);
if(!r)r= curRound;
switch(zt){
case nan:if(!(z.h&0x80000)){exceptions|= I_BIT;z.h|= 0x80000;}
case inf:case zro:return z;
case num:/*87:*/

if(ze>=1074)return fpack(zf,ze,zs,ROUND_OFF);
if(ze<=1020)xf.h= 0,xf.l= 1;
else{Octa oo;
xf= shiftRight(zf,1074-ze,1);
oo= shiftLeft(xf,1074-ze);
if(oo.l!=zf.l||oo.h!=zf.h)xf.l|= 1;

}
switch(r){
case ROUND_DOWN:if(zs=='-')xf= incr(xf,3);break;
case ROUND_UP:if(zs!='-')xf= incr(xf,3);
case ROUND_OFF:break;
case ROUND_NEAR:xf= incr(xf,xf.l&4?2:1);break;
}
xf.l&= 0xfffffffc;
if(ze>=1022)return fpack(shiftLeft(xf,1074-ze),ze,zs,ROUND_OFF);
if(xf.l)xf.h= 0x3ff00000,xf.l= 0;
if(zs=='-')xf.h|= SIGN_BIT;
return xf;

/*:87*/
;
}
  fprintf(stderr,
          "Error: this should not happen (fintegerize)\n");
  exit(1);
  return zero;
}

/*:86*//*88:*/

Octa fixit(z,r)
Octa z;
int r;
{
ftype zt;
int ze;
char zs;
Octa zf,o;
zt= funpack(z,&zf,&ze,&zs);
if(!r)r= curRound;
switch(zt){
case nan:case inf:exceptions|= I_BIT;return z;
case zro:return zero;
case num:if(funpack(fintegerize(z,r),&zf,&ze,&zs)==zro)return zero;
if(ze<=1076)o= shiftRight(zf,1076-ze,1);
else{
if(ze> 1085||(ze==1085&&(zf.h> 0x400000||
(zf.h==0x400000&&(zf.l||zs!='-')))))exceptions|= W_BIT;
if(ze>=1140)return zero;
o= shiftLeft(zf,ze-1076);
}
return(zs=='-'?ominus(zero,o):o);
}
  fprintf(stderr,
          "Error: this should not happen (fixit)\n");
  exit(1);
  return zero;
}

/*:88*//*89:*/

Octa floatit(Octa z, int r, int u, int p) {
  int e;
  char s;
  int t;

  exceptions= 0;
  if(!z.h&&!z.l)return zero;
  if(!r)r= curRound;
  if(!u&&(z.h&SIGN_BIT))s= '-',z= ominus(zero,z);else s= '+';
  e= 1076;
  while(z.h<0x400000)e--,z= shiftLeft(z,1);
  while(z.h>=0x800000){
    e++;
    t= z.l&1;
    z= shiftRight(z,1,1);
    z.l|= t;
  }
  if(p)/*90:*/{
    register int ex;register Tetra t;
    t= sfpack(z,e,s,r);
    ex= exceptions;
    sfunpack(t,&z,&e,&s);
    exceptions= ex;
  }
/*:90*/
  return fpack(z,e,s,r);
}

/*:89*//*91:*/

Octa froot(z,r)
Octa z;
int r;
{
ftype zt;
int ze;
char zs;
Octa x,xf,rf,zf;
register int xe,k;
if(!r)r= curRound;
zt= funpack(z,&zf,&ze,&zs);
if(zs=='-'&&zt!=zro)exceptions|= I_BIT,x= standardNaN;
else switch(zt){
case nan:if(!(z.h&0x80000))exceptions|= I_BIT,z.h|= 0x80000;
return z;
case inf:case zro:x= z;break;
case num:/*92:*/

xf.h= 0,xf.l= 2;
xe= (ze+0x3fe)>>1;
if(ze&1)zf= shiftLeft(zf,1);
rf.h= 0,rf.l= (zf.h>>22)-1;
for(k= 53;k;k--){
rf= shiftLeft(rf,2);xf= shiftLeft(xf,1);
if(k>=43)rf= incr(rf,(zf.h>>(2*(k-43)))&3);
else if(k>=27)rf= incr(rf,(zf.l>>(2*(k-27)))&3);
if((rf.l> xf.l&&rf.h>=xf.h)||rf.h> xf.h){
xf.l++;rf= ominus(rf,xf);xf.l++;
}
}
if(rf.h||rf.l)xf.l++;
return fpack(xf,xe,'+',r);

/*:92*/
;
}
if(zs=='-')x.h|= SIGN_BIT;
return x;
}

/*:91*//*93:*/

Octa fremstep(y,z,delta)
Octa y,z;
int delta;
{
ftype yt,zt;
int ye,ze;
char xs,ys,zs;
Octa x,xf,yf,zf;
register int xe,thresh,odd;
yt= funpack(y,&yf,&ye,&ys);
zt= funpack(z,&zf,&ze,&zs);
switch(4*yt+zt){
/*42:*/

case 4*nan+nan:if(!(y.h&0x80000))exceptions|= I_BIT;
case 4*zro+nan:case 4*num+nan:case 4*inf+nan:
if(!(z.h&0x80000))exceptions|= I_BIT,z.h|= 0x80000;
return z;
case 4*nan+zro:case 4*nan+num:case 4*nan+inf:
if(!(y.h&0x80000))exceptions|= I_BIT,y.h|= 0x80000;
return y;

/*:42*/
;
case 4*zro+zro:case 4*num+zro:case 4*inf+zro:
case 4*inf+num:case 4*inf+inf:x= standardNaN;
exceptions|= I_BIT;break;
case 4*zro+num:case 4*zro+inf:case 4*num+inf:return y;
case 4*num+num:/*94:*/

odd= 0;
thresh= ye-delta;
if(thresh<ze)thresh= ze;
while(ye>=thresh)/*95:*/

{
if(yf.h==zf.h&&yf.l==zf.l)goto zero_out;
if(yf.h<zf.h||(yf.h==zf.h&&yf.l<zf.l)){
if(ye==ze)goto try_complement;
ye--,yf= shiftLeft(yf,1);
}
yf= ominus(yf,zf);
if(ye==ze)odd= 1;
while(yf.h<0x400000)ye--,yf= shiftLeft(yf,1);
}

/*:95*/
;
if(ye>=ze){
exceptions|= E_BIT;return fpack(yf,ye,ys,ROUND_OFF);
}
if(ye<ze-1)return fpack(yf,ye,ys,ROUND_OFF);
yf= shiftRight(yf,1,1);
try_complement:xf= ominus(zf,yf),xe= ze,xs= '+'+'-'-ys;
if(xf.h> yf.h||(xf.h==yf.h&&(xf.l> yf.l||(xf.l==yf.l&&!odd))))
xf= yf,xs= ys;
while(xf.h<0x400000)xe--,xf= shiftLeft(xf,1);
return fpack(xf,xe,xs,ROUND_OFF);

/*:94*/
;
zero_out:x= zero;
}
if(ys=='-')x.h|= SIGN_BIT;
return x;
}
