/*
 * main.c -- main program
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "common.h"
#include "config.h"
#include "pipe.h"
#include "arith.h"
#include "memio.h"
#include "simio.h"


/**************************************************************/


#define BUF_SIZE	100


/**************************************************************/


static char *configFileName;
static char *progFileName;

static Octa curLoc;
static Octa curDat;
static Bool newChunk;
static char buffer[BUF_SIZE];
static FILE *progFile;
static Bool badAddress;

static int n, m;
static Octa bp = { -1, -1 };
static Octa tmp;

static Octa seven = { 0, 7 };


/**************************************************************/


static Bool undumpOcta(void) {
  int t0, t1, t2, t3;

  t0 = fgetc(progFile);
  if (t0 == EOF) {
    return false;
  }
  t1 = fgetc(progFile);
  if (t1 == EOF) {
    goto oops;
  }
  t2 = fgetc(progFile);
  if (t2 == EOF) {
    goto oops;
  }
  t3 = fgetc(progFile);
  if (t3 == EOF) {
    goto oops;
  }
  curDat.h = (t0 << 24) + (t1 << 16) + (t2 << 8) + (t3 << 0);
  t0 = fgetc(progFile);
  if (t0 == EOF) {
    goto oops;
  }
  t1 = fgetc(progFile);
  if (t1 == EOF) {
    goto oops;
  }
  t2 = fgetc(progFile);
  if (t2 == EOF) {
    goto oops;
  }
  t3 = fgetc(progFile);
  if (t3 == EOF) {
    goto oops;
  }
  curDat.l = (t0 << 24) + (t1 << 16) + (t2 << 8) + (t3 << 0);
  return true;
oops: ;
  fprintf(stderr,
          "Premature end of file on %s!\n",
          progFileName);
  return false;
}


static Octa readHex(char *p) {
  Octa val;
  int d[BUF_SIZE];
  int j, k;

  val.h = 0;
  val.l = 0;
  for (j = 0; ; j++) {
    if (p[j] >= '0' && p[j] <= '9') {
      d[j] = p[j] - '0';
    } else
    if (p[j] >= 'a' && p[j] <= 'f') {
      d[j] = p[j] - 'a' + 10;
    } else
    if (p[j] >= 'A' && p[j] <= 'F') {
      d[j] = p[j] - 'A' + 10;
    } else {
      break;
    }
  }
  p[j] = '\0';
  j--;
  for (k = 0; k <= j; k++) {
    if (k >= 8) {
      val.h += d[j - k] << (4 * k - 32);
    } else {
      val.l += d[j - k] << (4 * k);
    }
  }
  return val;
}


/**************************************************************/


int main(int argc, char *argv[]) {
  Fetch *newTail;
  int j;

  /* parse the command line <3 begin> */
  if (argc != 3) {
    fprintf(stderr,
            "Usage: %s <configfile> <progfile>\n",
            argv[0]);
    exit(-3);
  }
  configFileName = argv[1];
  progFileName = argv[2];
  /* parse the command line <3 end> */
  mmixConfig(configFileName);
  mmixInit();
  simioInit();
  /* input the program <4 begin> */
  if (strlen(progFileName) > 4 &&
      strcmp(progFileName + strlen(progFileName) - 4, ".mmb") == 0) {
    /* input an MMIX binary file <9 begin> */
    progFile = fopen(progFileName, "rb");
    if (progFile == NULL) {
      fprintf(stderr,
              "Panic: Cannot open MMIX binary file %s!\n",
              progFileName);
      exit(-3);
    }
    while (1) {
      if (!undumpOcta()) {
        break;
      }
      newChunk = true;
      curLoc = curDat;
      if (curLoc.h & 0x9FFFFFFF) {
        badAddress = true;
      } else {
        badAddress = false;
        /* apply trivial mapping function for each segment */
        curLoc.h >>= 29;
      }
      /* input consecutive octabytes beginning at curLoc <11 begin> */
      while (1) {
        if (!undumpOcta()) {
          fprintf(stderr,
                  "Unexpected end of file on %s!\n",
                  progFileName);
          break;
        }
        if (curDat.h == 0 && curDat.l == 0) {
          break;
        }
        if (badAddress) {
          fprintf(stderr,
                  "Panic: Unsupported virtual address #%08X%08X!\n",
                  curLoc.h, curLoc.l);
          exit(-5);
        }
        if (newChunk) {
          memWrite(curLoc, curDat);
        } else {
          memHash[lastHash].chunk[(curLoc.l & 0xFFFF) >> 3] = curDat;
        }
        curLoc.l += 8;
        if ((curLoc.l & 0xFFF8) != 0) {
          newChunk = false;
        } else {
          newChunk = true;
          if ((curLoc.l & 0xFFFF0000) == 0) {
            badAddress = true;
            curLoc.h = (curLoc.h << 29) + 1;
          }
        }
      }
      /* input consecutive octabytes beginning at curLoc <11 end> */
    }
    /* set up the canned environment <12 begin> */
    if (curLoc.h != 3) {
      fprintf(stderr,
              "Panic: MMIX binary file did not set up the stack!\n");
      exit(-6);
    }
    instPtr.o = memRead(incr(curLoc, -8 * 14));  /* Main */
    instPtr.p = NULL;
    curLoc.h = 0x60000000;
    g[255].o = incr(curLoc, -8);  /* place to UNSAVE */
    curDat.l = 0x90;
    if (memRead(curDat).h) {
      instPtr.o = curDat;  /* start at #90 if nonzero */
    }
    head->inst = (UNSAVE << 24) + 255;  /* prefetch a fabricated command */
    tail--;
    head->loc = incr(instPtr.o, -4);  /* in case the UNSAVE is interrupted */
    g[rT].o.h = 0x80000005;
    g[rTT].o.h = 0x80000006;
    curDat.h = (RESUME << 24) + 1;
    curDat.l = 0;
    curLoc.h = 5;
    curLoc.l = 0;
    memWrite(curLoc, curDat);  /* primitive trap handler */
    curDat.l = curDat.h;
    curDat.h = (NEGI << 24) + (255 << 16) + 1;
    curLoc.h = 6;
    curLoc.l = 8;
    memWrite(curLoc, curDat);  /* primitive dynamic trap handler */
    curDat.h = (GET << 24) + rQ;
    curDat.l = (PUTI << 24) + (rQ << 16);
    curLoc.l = 0;
    memWrite(curLoc, curDat);  /* more of primitive dynamic trap handler */
    curDat.h = 0;  /* generate a PTE with rwx permission */
    curDat.l = 7;
    curLoc.h = 4;  /* beginning of skeleton page table */
    memWrite(curLoc, curDat);  /* PTE for the text segment */
    ITcache->set[0][0].tag = zero;
    ITcache->set[0][0].data[0] = curDat;  /* prime the IT cache */
    curDat.l = 6;  /* PTE with read and write permission only */
    curDat.h = 1;
    curLoc.l = 3 << 13;
    memWrite(curLoc, curDat);  /* PTE for the data segment */
    curDat.h = 2;
    curLoc.l = 6 << 13;
    memWrite(curLoc, curDat);  /* PTE for the pool segment */
    curDat.h = 3;
    curLoc.l = 9 << 13;
    memWrite(curLoc, curDat);  /* PTE for the stack segment */
    g[rK].o = negOne;  /* enable all interrupts */
    g[rV].o.h = 0x369C2004;
    pageBad = false;
    pageR = 4 << (32 - 13);
    pageS = 32;
    pageMask.l = 0xFFFFFFFF;
    pageB[1] = 3;
    pageB[2] = 6;
    pageB[3] = 9;
    pageB[4] = 12;
    /* set up the canned environment <12 end> */
    /* input an MMIX binary file <9 end> */
  } else {
    /* input a rudimentary hexadecimal file <6 begin> */
    progFile = fopen(progFileName, "r");
    if (progFile == NULL) {
      fprintf(stderr,
              "Panic: Cannot open MMIX hexadecimal file %s!\n",
              progFileName);
      exit(-3);
    }
    newChunk = true;
    while (1) {
      if (fgets(buffer, BUF_SIZE, progFile) == NULL) {
        break;
      }
      if (buffer[strlen(buffer) - 1] != '\n') {
        fprintf(stderr,
                "Panic: Hexadecimal file line too long: '%s...'!\n",
                buffer);
        exit(-3);
      }
      if (buffer[12] == ':') {
        /* change the current location <7 begin> */
        if (sscanf(buffer, "%4x%8x", &curLoc.h, &curLoc.l) != 2) {
          fprintf(stderr,
                  "Panic: Improper hexadecimal file location: '%s'!\n",
                  buffer);
          exit(-3);
        }
        newChunk = true;
        /* change the current location <7 end> */
      } else
      if (buffer[0] == ' ') {
        /* read an octabyte and advance curLoc <8 begin> */
        if (sscanf(buffer + 1, "%8x%8x", &curDat.h, &curDat.l) != 2) {
          fprintf(stderr,
                  "Panic: Improper hexadecimal file data: '%s'!\n",
                  buffer);
          exit(-3);
        }
        if (newChunk) {
          memWrite(curLoc, curDat);
        } else {
          memHash[lastHash].chunk[(curLoc.l & 0xFFFF) >> 3] = curDat;
        }
        curLoc.l += 8;
        if ((curLoc.l & 0xFFF8) != 0) {
          newChunk = false;
        } else {
          newChunk = true;
          if ((curLoc.l & 0xFFFF0000) == 0) {
            curLoc.h++;
          }
        }
        /* read an octabyte and advance curLoc <8 end> */
      } else {
        fprintf(stderr,
                "Panic: Improper hexadecimal file line: '%s'!\n",
                buffer);
        exit(-3);
      }
    }
    /* input a rudimentary hexadecimal file <6 end> */
  }
  fclose(progFile);
  /* input the program <4 end> */
  /* run the simulation interactively <13 begin> */
  while (1) {
    printf("MMIX > ");
    fflush(stdout);
    if (fgets(buffer, BUF_SIZE, stdin) == NULL) {
      printf("\n");
      goto done;
    }
    switch (buffer[0]) {
      case '\n':
        continue;
      case 'q':
      case 'x':
        goto done;
      case 'h':
      case '?':
        printf("The interactive commands are as follows:\n");
        printf(" <n> to run for n cycles\n");
        printf(" @<x> to take next instruction from location x\n");
        printf(" b<x> to pause when location x is fetched\n");
        printf(" v<x> to print specified diagnostics when running;\n");
        printf("    x=1[insts enter/leave pipe]+2[whole pipeline each cycle]+\n");
        printf("      4[coroutine activations]+8[coroutine scheduling]+\n");
        printf("      10[uninitialized read]+20[online I/O read]+\n");
        printf("      40[I/O read/write]+80[branch prediction details]+\n");
        printf("      100[invalid cache blocks displayed too]\n");
        printf(" -<n> to deissue n instructions\n");
        printf(" l<n> to print current value of local register n\n");
        printf(" g<n> to print current value of global register n\n");
        printf(" m<x> to print current value of memory address x\n");
        printf(" f<x> to insert instruction x into the fetch buffer\n");
        printf(" i<n> to initiate a timer interrupt after n cycles\n");
        printf(" IT, DT, I, D, or S to print current cache contents\n");
        printf(" D* or S* to print dirty blocks of a cache\n");
        printf(" p to print current pipeline contents\n");
        printf(" s to print current stats\n");
        printf(" h to print this message\n");
        printf(" q to exit\n");
        printf("(Here <n> is a decimal integer, <x> is hexadecimal.)\n");
        continue;
      case '0':
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
        if (sscanf(buffer, "%d", &n) != 1) {
          goto whatSay;
        }
        printf("Running %d at time %d", n, ticks.l);
        if (bp.h == (Tetra) -1 && bp.l == (Tetra) -1) {
          printf("\n");
        } else {
          printf(" with breakpoint %08X%08X\n", bp.h, bp.l);
        }
        mmixRun(n, bp);
        continue;
      case '@':
        instPtr.o = readHex(buffer + 1);
        instPtr.p = NULL;
        continue;
      case 'b':
        bp = readHex(buffer + 1);
        continue;
      case 'v':
        verbose = readHex(buffer + 1).l;
        continue;
      case '-':
        if (sscanf(buffer + 1, "%d", &n) != 1 || n < 0) {
          goto whatSay;
        }
        if (cool <= hot) {
          m = hot - cool;
        } else {
          m = (hot - reorderBot) + 1 + (reorderTop - cool);
        }
        if (n > m) {
          deissues = m;
        } else {
          deissues = n;
        }
        continue;
      case 'l':
        if (sscanf(buffer + 1, "%d", &n) != 1 || n < 0) {
          goto whatSay;
        }
        if (n >= lringSize) {
          goto whatSay;
        }
        printf("  l[%d]=%08X%08X\n", n, l[n].o.h, l[n].o.l);
        continue;
      case 'm':
        tmp = memRead(readHex(buffer + 1));
        printf("  m[%s]=%08X%08X\n", buffer + 1, tmp.h, tmp.l);
        continue;
      case 'g':
        if (sscanf(buffer + 1, "%d", &n) != 1 || n < 0) {
          goto whatSay;
        }
        if (n >= 256) {
          goto whatSay;
        }
        if (n == rO || n == rS) {
          /* these two are not kept up to date in the g array */
          if (hot == cool) {
            /* pipeline empty */
            g[rO].o = shiftLeft(coolO, 3);
            g[rS].o = shiftLeft(coolS, 3);
          } else {
            /* pipeline not empty */
            g[rO].o = shiftLeft(hot->curO, 3);
            g[rS].o = shiftLeft(hot->curS, 3);
          }
        }
        printf("  g[%d]=%08X%08X\n", n, g[n].o.h, g[n].o.l);
        continue;
      case 'I':
        printCache(buffer[1] == 'T' ? ITcache : Icache, false);
        continue;
      case 'D':
        printCache(buffer[1] == 'T' ? DTcache : Dcache, buffer[1] == '*');
        continue;
      case 'S':
        printCache(Scache, buffer[1] == '*');
        continue;
      case 'p':
        printPipe();
        printLocks();
        continue;
      case 's':
        printStats();
        continue;
      case 'i':
        if (sscanf(buffer + 1, "%d", &n) == 1) {
          g[rI].o = incr(zero, n);
        }
        continue;
      case 'f':
        tmp = readHex(buffer + 1);
        if (tail == fetchBot) {
          newTail = fetchTop;
        } else {
          newTail = tail - 1;
        }
        if (newTail == head) {
          printf("Sorry, the fetch buffer is full!\n");
        } else {
          tail->loc = instPtr.o;
          tail->inst = tmp.l;
          tail->interrupt = 0;
          tail->noted = false;
          tail = newTail;
        }
        continue;
      case 'd':
        if (ticks.l) {
          printf("Sorry, I disable ITcache and DTcache only at the beginning!\n");
        } else {
          ITcache->set[0][0].tag = zero;
          ITcache->set[0][0].data[0] = seven;
          DTcache->set[0][0].tag = zero;
          DTcache->set[0][0].data[0] = seven;
          g[rK].o = negOne;
          pageBad = false;
          pageMask = negOne;
          instPtr.p = (SpecNode *) 1;
        }
        continue;
      case 'k':
        for (j = 0; j < funcUnitCount; j++) {
          printf("unit %s %d\n", funcUnit[j].name, funcUnit[j].st);
        }
        continue;
      default:
whatSay: ;
        printf("Eh? Sorry, I don't understand. (Type h for help)\n");
        continue;
    }
  }
done: ;
  /* run the simulation interactively <13 end> */
  printf("Simulation ended at time %d.\n", ticks.l);
  printStats();
  return 0;
}
