/*
 * common.h -- common definitions
 */


#ifndef _COMMON_H_
#define _COMMON_H_


#define VERSION		1		/* version of the MMIX architecture */
#define SUBVERSION	0		/* secondary byte of version number */
#define SUBSUBVERSION	2		/* further qualification of version */

#define K		1024		/* Kilo */
#define M		(K * K)		/* Mega */

#define MEMSIZE_MIN	(1 * M)		/* minimum memory size */
#define MEMSIZE_MAX	(16 * M)	/* maximum memory size */
#define MEMSIZE_DEF	(4 * M)		/* default memory size */
#define MEM_BASE_H	0x00000000	/* phys base addr of mem, hi part */
#define MEM_BASE_L	0x00000000	/* ditto, lo part, must be 0 */

#define ROM_SIZE	(64 * K)	/* ROM size */
#define ROM_BASE_H	0x00008000	/* phys base addr of ROM, hi part */
#define ROM_BASE_L	0x00000000	/* ditto, lo part, must be 0 */

#define DEV_TYPE_MASK	0xFFFF0000	/* device type bits in hi addr */
#define DEV_NUM_MASK	0x0000FFFF	/* device number bits in hi addr */

#define TERM_BASE_H	0x00010000	/* terminal base addr, hi part */
#define TERM_BASE_L	0x00000000	/* ditto, lo part, must be 0 */
#define DISK_BASE_H	0x00020000	/* disk base addr, hi part */
#define DISK_BASE_L	0x00000000	/* ditto, lo part, must be 0 */
#define GRAPH_BASE_H	0x00030000	/* graphics base addr, hi part */
#define GRAPH_BASE_L	0x00000000	/* ditto, lo part, must be 0 */

#define NTERMS_MIN	0		/* minimum number of terminals */
#define NTERMS_MAX	4		/* maximum number of terminals */
#define NTERMS_DEF	1		/* default number of terminals */

#define DEFAULT_CFG	"default.cfg"	/* default configuration file name */

#define CYCS_PER_MSEC	500		/* average execution speed */

#define SIGN(n)		(((tetra) 1) << ((n) - 1))	/* n = 1..32 */
#define MASK(n)		((((tetra) 1) << (n)) - 1)	/* n = 0..31 */


typedef enum { false, true } bool;	/* truth values */

typedef unsigned char U8;
typedef signed char S8;
typedef unsigned short U16;
typedef signed short S16;
typedef unsigned int U32;
typedef signed int S32;

typedef U8 byte;			/* 8-bit quantities */
typedef U16 wyde;			/* 16-bit quantities */
typedef U32 tetra;			/* 32-bit quantities */
typedef struct {
  tetra h;
  tetra l;
} octa;					/* 64-bit quantities */


#endif /* _COMMON_H_ */
