/*
 * cpu.h -- CPU simulation
 */


#ifndef _CPU_H_
#define _CPU_H_


#define random my_random
#define fsqrt my_fsqrt
#define div my_div
#define sync my_sync

#define issue_bit (1<<0)
#define pipe_bit (1<<1)
#define coroutine_bit (1<<2)
#define schedule_bit (1<<3)
#define show_pred_bit (1<<7)
#define show_wholecache_bit (1<<8)

#define POWER_FAILURE (1<<0)
#define PARITY_ERROR (1<<1)
#define NONEXISTENT_MEMORY (1<<2)
#define REBOOT_SIGNAL (1<<4)
#define INTERVAL_TIMEOUT (1<<7)

#define ticks g[rC].o

#define max_stage 99
#define vanish 98
#define flush_to_mem 97
#define flush_to_S 96
#define fill_from_mem 95
#define fill_from_S 94
#define fill_from_virt 93
#define write_from_wbuf 92
#define cleanup 91

#define WRITE_BACK 1
#define WRITE_ALLOC 2


typedef struct coroutine_struct {
  char *name;
  int stage;
  struct coroutine_struct *next;
  struct coroutine_struct **lockloc;
  struct control_struct *ctl;
} coroutine;

typedef coroutine *lockvar;

typedef struct {
  octa o;
  struct specnode_struct *p;
} spec;

typedef struct specnode_struct {
  octa o;
  bool known;
  octa addr;
  struct specnode_struct *up, *down;
} specnode;

typedef enum {
  TRAP, FCMP, FUN, FEQL, FADD, FIX, FSUB, FIXU,
  FLOT, FLOTI, FLOTU, FLOTUI, SFLOT, SFLOTI, SFLOTU, SFLOTUI,
  FMUL, FCMPE, FUNE, FEQLE, FDIV, FSQRT, FREM, FINT,
  MUL, MULI, MULU, MULUI, DIV, DIVI, DIVU, DIVUI,
  ADD, ADDI, ADDU, ADDUI, SUB, SUBI, SUBU, SUBUI,
  IIADDU, IIADDUI, IVADDU, IVADDUI, VIIIADDU, VIIIADDUI, XVIADDU, XVIADDUI,
  CMP, CMPI, CMPU, CMPUI, NEG, NEGI, NEGU, NEGUI,
  SL, SLI, SLU, SLUI, SR, SRI, SRU, SRUI,
  BN, BNB, BZ, BZB, BP, BPB, BOD, BODB,
  BNN, BNNB, BNZ, BNZB, BNP, BNPB, BEV, BEVB,
  PBN, PBNB, PBZ, PBZB, PBP, PBPB, PBOD, PBODB,
  PBNN, PBNNB, PBNZ, PBNZB, PBNP, PBNPB, PBEV, PBEVB,
  CSN, CSNI, CSZ, CSZI, CSP, CSPI, CSOD, CSODI,
  CSNN, CSNNI, CSNZ, CSNZI, CSNP, CSNPI, CSEV, CSEVI,
  ZSN, ZSNI, ZSZ, ZSZI, ZSP, ZSPI, ZSOD, ZSODI,
  ZSNN, ZSNNI, ZSNZ, ZSNZI, ZSNP, ZSNPI, ZSEV, ZSEVI,
  LDB, LDBI, LDBU, LDBUI, LDW, LDWI, LDWU, LDWUI,
  LDT, LDTI, LDTU, LDTUI, LDO, LDOI, LDOU, LDOUI,
  LDSF, LDSFI, LDHT, LDHTI, CSWAP, CSWAPI, LDUNC, LDUNCI,
  LDVTS, LDVTSI, PRELD, PRELDI, PREGO, PREGOI, GO, GOI,
  STB, STBI, STBU, STBUI, STW, STWI, STWU, STWUI,
  STT, STTI, STTU, STTUI, STO, STOI, STOU, STOUI,
  STSF, STSFI, STHT, STHTI, STCO, STCOI, STUNC, STUNCI,
  SYNCD, SYNCDI, PREST, PRESTI, SYNCID, SYNCIDI, PUSHGO, PUSHGOI,
  OR, ORI, ORN, ORNI, NOR, NORI, XOR, XORI,
  AND, ANDI, ANDN, ANDNI, NAND, NANDI, NXOR, NXORI,
  BDIF, BDIFI, WDIF, WDIFI, TDIF, TDIFI, ODIF, ODIFI,
  MUX, MUXI, SADD, SADDI, MOR, MORI, MXOR, MXORI,
  SETH, SETMH, SETML, SETL, INCH, INCMH, INCML, INCL,
  ORH, ORMH, ORML, ORL, ANDNH, ANDNMH, ANDNML, ANDNL,
  JMP, JMPB, PUSHJ, PUSHJB, GETA, GETAB, PUT, PUTI,
  POP, RESUME, SAVE, UNSAVE, SYNC, SWYM, GET, TRIP
} mmix_opcode;

#define max_pipe_op feps
#define max_real_command trip

typedef enum {
  mul0,
  mul1,
  mul2,
  mul3,
  mul4,
  mul5,
  mul6,
  mul7,
  mul8,
  div,
  sh,
  mux,
  sadd,
  mor,
  fadd,
  fmul,
  fdiv,
  fsqrt,
  fint,
  fix,
  flot,
  feps,
  fcmp,
  funeq,
  fsub,
  frem,
  mul,
  mulu,
  divu,
  add,
  addu,
  sub,
  subu,
  set,
  or,
  orn,
  nor,
  and,
  andn,
  nand,
  xor,
  nxor,
  shlu,
  shru,
  shl,
  shr,
  cmp,
  cmpu,
  bdif,
  wdif,
  tdif,
  odif,
  zset,
  cset,
  get,
  put,
  ld,
  ldptp,
  ldpte,
  ldunc,
  ldvts,
  preld,
  prest,
  st,
  syncd,
  syncid,
  pst,
  stunc,
  cswap,
  br,
  pbr,
  pushj,
  go,
  prego,
  pushgo,
  pop,
  resume,
  save,
  unsave,
  sync,
  jmp,
  noop,
  trap,
  trip,
  incgamma,
  decgamma,
  incrl,
  sav,
  unsav,
  resum
} internal_opcode;

typedef struct control_struct {
  octa loc;
  mmix_opcode op;
  unsigned char xx, yy, zz;
  spec y, z, b, ra;
  specnode x, a, go, rl;
  coroutine *owner;
  internal_opcode i;
  int state;
  bool usage;
  bool need_b;
  bool need_ra;
  bool ren_x;
  bool mem_x;
  bool ren_a;
  bool set_l;
  bool interim;
  unsigned int arith_exc;
  unsigned int hist;
  int denin, denout;
  octa cur_O, cur_S;
  unsigned int interrupt;
  void *ptr_a, *ptr_b, *ptr_c;
} control;

typedef struct {
  octa loc;
  tetra inst;
  unsigned int interrupt;
  bool noted;
  unsigned int hist;
} fetch;

typedef struct func_struct {
  char name[16];
  tetra ops[8];
  int k;
  coroutine *co;
} func;

typedef enum { random, serial, pseudo_lru, lru } replace_policy;

typedef struct {
  octa tag;
  char *dirty;
  octa *data;
  int rank;
} cacheblock;

typedef cacheblock *cacheset;

typedef struct {
  int a, b, c, g, v;

  int aa, bb, cc, gg, vv;

  int tagmask;
  replace_policy repl, vrepl;
  int mode;
  int access_time;
  int copy_in_time;
  int copy_out_time;
  cacheset *set;
  cacheset victim;
  coroutine filler;
  control filler_ctl;
  coroutine flusher;

  control flusher_ctl;
  cacheblock inbuf;
  cacheblock outbuf;
  lockvar lock;
  lockvar fill_lock;
  int ports;
  coroutine *reader;

  char *name;
} cache;

//typedef struct {
//  tetra tag;
//  octa *chunk;
//} chunknode;

typedef struct {
  octa o;
  octa addr;
  tetra stamp;
  internal_opcode i;
} write_node;

//typedef enum {
//  Halt, Fopen, Fclose, Fread, Fgets, Fgetws,
//  Fwrite, Fputs, Fputws, Fseek, Ftell
//} sys_call;


extern int verbose;

extern int ring_size;
extern coroutine *ring;
extern int cur_time;

extern int fetch_max, dispatch_max, peekahead, commit_max;

extern control *reorder_bot, *reorder_top;

extern control *hot, *cool;
extern control *old_hot;
extern int deissues;

extern int *dispatch_stat;

extern bool security_disabled;

extern fetch *fetch_bot, *fetch_top;

extern fetch *head, *tail;

extern func *funit;
extern int funit_count;

extern specnode g[256];
extern specnode *l;
extern int lring_size;

extern int max_rename_regs, max_mem_slots;
extern int rename_regs, mem_slots;

extern octa cool_O, cool_S;

extern specnode mem;

#define pipe_limit 90
extern unsigned char pipe_seq[max_pipe_op + 1][pipe_limit + 1];

extern int bp_a, bp_b, bp_c, bp_n;
extern char *bp_table;

extern cache *Icache, *Dcache, *Scache, *ITcache, *DTcache;

//extern int mem_chunks;
//extern int mem_chunks_max;
//extern int hash_prime;
//extern chunknode *mem_hash;

//extern int last_h;

extern int mem_addr_time;
extern int bus_words;
extern int mem_read_time;
extern int mem_write_time;
extern lockvar mem_lock;

extern bool no_hardware_PT;

extern write_node *wbuf_bot, *wbuf_top;

extern write_node *write_head, *write_tail;

extern lockvar wbuf_lock;
extern int holding_time;
extern lockvar speed_lock;

extern spec inst_ptr;
extern octa *fetched;

extern int frem_max;
extern int denin_penalty, denout_penalty;


void print_locks(void);
void print_stats(void);
void print_cache(cache *, bool);
void clean_block(cache *, cacheblock *);
void zap_cache(cache *);
void print_pipe(void);

octa cpuGetGR(int regnum);
void cpuSetGR(int regnum, octa value);
octa cpuGetLR(int regnum);
void cpuSetLR(int regnum, octa value);
octa cpuGetDR(int regnum);
void cpuSetDR(int regnum, octa value);

bool cpuFetch(tetra instr);

bool cpuTestIP(void);
octa cpuGetIP(void);
void cpuSetIP(octa addr);

bool cpuTestBreak(void);
octa cpuGetBreak(void);
void cpuSetBreak(octa addr);
void cpuResetBreak(void);

void cpuStep(void);
void cpuRun(void);
void cpuHalt(void);

void cpuReset(void);
void cpuInit(void);
void cpuExit(void);


#endif /* _CPU_H_ */
